;;;-----------------------------------------------------------------------------------
;;; name       : the-net.lisp
;;; description: relaxation-net consists of two kinds of nodes: function-nodes 
;;;              and value-nodes. Value-nodes contain values of some kind (numbers,
;;;              sets, lists etc.). These values are changed by the 
;;;              relaxation-functions of the connected function-nodes. The 
;;;              relaxation-functions are called (activated), when the value of a 
;;;              connected value-node has changed. This process - called relaxation -
;;;              continues until a write to a value-node fails or an equilibrium
;;;              is reached (no more activated function-nodes)
;;; notes      : locking muss ermoeglicht werden. Der Konstruktor make-net muss besser
;;;              werden. 
;;; contact    : me (Michael Trowe)
;;; copyright  : mine, mine, it's all mine! Harharharhar
;;; history    : 
;;; contents   :
;;;-----------------------------------------------------------------------------------
(in-package nc)

(defvar *number-of-relaxing-processes* 20)

(defvar *relax-net* ())

(defclass logg-obj () 
  ((logg-table :accessor logg-table :initform (make-hash-table))))
()

(defparameter *my-trace-channel* (make-trace-channel :net-trace))

(defpargeneric stopped :now (net))

(defclass relaxation-net (tig chaco-strategy logg-obj)
  ((name :accessor net-name
	 :initarg :name
	 :initform ())
   (partitions :accessor partitions :initform ())))
  

(defclass partition (ncl-object logg-obj)
   ((value-nodes :accessor value-nodes
		:initform ())
   (function-nodes :accessor function-nodes
		   :initform ())
   (stopped :accessor stopped
	    :initform ())
   (acknowledge-count :accessor acknowledge-count :initform 0))
   (:metaclass mobile-object-class))

(defpargeneric stop-relax :past (net))


(defclass net-node (task logg-obj)
  ((name :accessor node-name
	 :initarg :name
	 :initform ())
   (net :accessor net :moving-behavior :stay))
  (:metaclass mobile-object-class))


(defclass function-node (net-node active-object)
  ((connected-tasks :accessor value-nodes
		:initform ()
		:moving-behavior :stay)
   (relaxation-function :accessor relax-func
			:initarg :relax-func
			:moving-behavior :follow)
   (parent :initform() :accessor parent)
   (acknowledge-count :initform 0 :accessor acknowledge-count)
   (relaxing :initform nil :accessor relaxing))
  (:metaclass mobile-object-class))


(defpargeneric test-activation :past (f-node))


(defclass value-node (net-node)
  ((content :accessor content
	    :initarg :content
	    :moving-behavior :follow)
   (connected-tasks :accessor function-nodes
		   :initform ()
		   :moving-behavior :stay)
   (v-lock :accessor v-lock :initform (mp:make-process-lock) 
	   :moving-behavior :ignore))
  (:metaclass  mobile-object-class))

(defclass net-scheduler (standard-scheduler autonomous-object logg-obj)
  ((node-queue :accessor node-queue :initform (make-instance 'lockable-queue))
   (full :accessor full :initform ())))

(defmethod schedule ((sched net-scheduler) node message)
  (if (eq (second message) 'relax) 
      (if (relaxing node)		
	  ;; if someone sent another relax message to the node that wasn't
	  ;; yet applied you don't need to apply this message.
	  (send-past *current-actor* node 'acknowledge)
	(progn (setf (relaxing node) t)
	       (enqueue (mail-queue node) message)
	       (unless (active-p node)
		 (setf (active-p node) t)
		 (enqueue (node-queue sched) node))))
    (progn (enqueue (mail-queue node) message)
	   (unless (active-p node)
	     (setf (active-p node) t)
	     (enqueue (node-queue sched) node)))))
	   


(defmethod active-loop ((sched net-scheduler))
  (loop
    (when (empty-queue-p (node-queue sched))
      (logg sched 'waiting (multiple-value-bind (sec min hour) 
			       (get-decoded-time) 
			     (format nil "~a:~a,~a Uhr" hour min sec)))
      (mp:process-wait "no nodes to relax" 
		       #'(lambda () (not (empty-queue-p (node-queue sched)))))
      (logg sched 'resuming (multiple-value-bind (sec min hour) 
				(get-decoded-time) 
			      (format nil "~a:~a,~a Uhr" hour min sec))))
    (let ((node (dequeue (node-queue sched))))
      (active-body node)
      (when (active-test node)
	(enqueue (node-queue sched) node)))))



(let ((schedulers  ())
      (i 0))
  (defun get-net-scheduler (f-node)
    (let ((sched (prog1 (nth i schedulers) (setf i (mod (1+ i) *number-of-relaxing-processes*)))))
      (if (full sched)
	  (get-net-scheduler f-node)
	(progn (when (border-node f-node)
		 (setf (full sched) t))
	       sched))))
  
  (defun reset-net-scheduler ()
    (loop for sched in schedulers do (kill sched))
    (setf schedulers (loop repeat *number-of-relaxing-processes* 
			 collect (make-instance 'net-scheduler)))
    ()))
  




(defpargeneric acknowledge :past (node))

(defpargeneric write-value :now (v-node writer value))


(defpargeneric read-value :now (v-node))

(defpargeneric relax :past (f-node))


(defmethod (setf tasks) :before ((tasks list) (net relaxation-net))
    (loop with partitioning = ()
      for task in tasks
      for part = (find task partitioning :test  #'(lambda (task part) 
						    (eq (location task)
							(location (first part)))))
      if part
      do (push task (cdr part))
      else 
      do (push (list (ensure-partition net (location task))) partitioning)
      finally (loop for part in partitioning
		  do (setf (tasks (first part)) (rest part)))))
 

(defmethod (setf tasks) ((tasks list) (part partition))
  (loop for task in tasks
      do (add-task part task))
  tasks)

(defmethod tasks ((partition partition))
  (append (value-nodes partition) (function-nodes partition)))

(defmethod add-task ((part partition) (task function-node))
  (push task (function-nodes part)))

(defmethod  add-task ((part partition) (task value-node))
  (push task (value-nodes part)))


(defmethod add-task ((net relaxation-net) (task function-node))
  (let ((part (ensure-partition net (location task))))
  (push task (function-nodes part))))

(defmethod  add-task ((net relaxation-net) (task value-node))
  (let ((part (ensure-partition net (location task))))
    (push task (value-nodes part))))

(defmethod value-nodes ((net relaxation-net))
  (loop for part in (partitions net)
      append (value-nodes part)))

(defmethod function-nodes ((net relaxation-net))
  (loop for part in (partitions net)
      append (function-nodes part)))

(defmethod ensure-partition ((net relaxation-net) location)
  (or (find location (partitions net) :test #'(lambda (loc part) (eq loc (location part))))
      (let ((part (make-instance 'partition :location location)))
	(push part (partitions net))
	part)))
						      

(defun make-net (f-descrs v-descrs connectivity)
  (let ((net (make-graph 'relaxation-net 
			 (append (loop for f in f-descrs 
				     collect (cons 'function-node f))
				 (loop for v in v-descrs
				     collect (cons 'value-node v)))
			 (append (loop with f-con = (make-list (length f-descrs) 
							       :initial-element 0)
				     for row in connectivity
				     collect (append f-con row))
				 (loop with v-con = (make-list (length v-descrs) 
							       :initial-element 0)
				     for x from 0 upto (1- (length v-descrs))
				     for column = (mapcar #'(lambda (row) 
							      (nth x row))
							  connectivity)
				     collect (append column v-con))))))
    (loop for task in (tasks net)
	do (setf (net task) net))
    net))


(defmacro defnet (name v-nodes f-nodes)
  `(progn (reset-net-scheduler)
	  (loop for space in (spaces  *manager*)
			     do (remote-eval space '(reset-net-scheduler)))
       (let ((net (make-graph 'relaxation-net 
			      ',(compute-nodes v-nodes f-nodes)
			      ',(compute-connectivity v-nodes f-nodes)
			      :name ',name)))
	 (loop for part in (partitions net)
	     do (init-part part net))
	 net)))

(defun compute-nodes (v-nodes f-nodes)
  (loop for node in (append (mapcar #'make-v-node v-nodes)
			    (mapcar #'(lambda (f-node)
					(make-f-node f-node v-nodes))
				    f-nodes))
      collect (list node 1)))

(defun make-v-node (node-descr)
  `(value-node :name ,@node-descr))

(defun make-f-node (node-descr v-nodes)
  `(function-node :name ,(first node-descr)
		  :relax-func ,(make-relax-func (rest node-descr) v-nodes)))

(defun make-relax-func (func-call v-nodes)
  (let ((result (gentemp "result" "NC"))
	(args (v-node-order (rest func-call) v-nodes)))
   `(lambda ,args
     (let ((,result (funcall ,@func-call)))
       (list ,@(unorder (rest func-call) args result))))))

(defun v-node-order (arg-nodes v-nodes)
  (loop for v-node in v-nodes
      when (member (first v-node) arg-nodes)
      collect (first v-node)))

(defun unorder (arg-nodes args result)
  (loop for arg in args
      collect `(nth ,(position arg arg-nodes) ,result)))

(defun compute-connectivity (v-nodes f-nodes)
  (let* ((v-length (length v-nodes))
	 (f-length (length f-nodes))
	 (edges  (make-array (+ v-length f-length) :initial-element ())))
    (loop for i from 0 upto (1- v-length)
	do (loop for j from v-length upto (+ f-length v-length)
	       for f-node in f-nodes
	       when (member (first (nth i v-nodes))
			    (cddr f-node))
	       do (progn (setf (aref edges i) (cons j (aref edges i)))
			 (setf (aref edges j) (cons i (aref edges j))))))
    edges))


(defpargeneric init-part :past (part net))

(defmethod init-part ((part partition) (net relaxation-net))
  (loop for task in (tasks part)
      do (setf (net task) net))
  (loop for f in (function-nodes part)
      do (setf (scheduler f) (get-net-scheduler f))))


		



;;; write a value to the node and send a relax-message to the
;;; connected function-nodes.
;;; The acknowledge stuff is for termination detection.
;;; write-value returns the number of function-nodes relax is called on. 
(defmethod write-value ((node value-node) writer value)
  ;;(format-trace *my-trace-channel* "writing variable ~a for constraint ~a.~%" 
  ;;	(node-name node) (node-name *current-actor*))
  (cond ((not (value-ok node value))
	 ;(mp:process-unlock (v-lock node) *current-actor*) 
	 (stop-relax (net node) 0))
	((value-new node value)
	 (setf (content node) value)
	 ;(mp:process-unlock (v-lock node) *current-actor*)
	 (loop for f-node in (function-nodes node)
	     unless (eq f-node writer)
	     sum (prog1 1 
		   (relax f-node))))
	(t ;(mp:process-unlock (v-lock node) *current-actor*) 
	   0)))
 


(defmethod read-value ((node value-node))
  ;;(format-trace *my-trace-channel* "reading variable ~a for constraint ~a.~%" 
  ;;	(name node) (name *current-actor*))
  ;(mp:process-lock (v-lock node) *current-actor*)
  (content node))


(defmethod acknowledge ((node function-node))
  (decf (acknowledge-count node))
  (when (and (parent node) (= (acknowledge-count node) 0))
    (acknowledge (parent node))
    (setf (parent node) ())))


  
;;; applying the relaxation function and doing some termination detection. 
(defmethod relax ((f-node function-node))
  (if (= (acknowledge-count f-node) 0) 
      (setf (parent f-node) *caller*)
    (acknowledge *caller*))
  ;;(format-trace *my-trace-channel* "relaxing node ~a.~%" (node-name f-node))
  (loop for v-node in (value-nodes f-node)
      for value in (prog1 (apply (relax-func f-node) 
				 (loop for v-node in (value-nodes f-node)
				     collect (read-value v-node)))
		     ;(loop for x from 1 to 1000 do (* x x))
		     (setf (relaxing f-node) nil))
      do (incf (acknowledge-count f-node) (write-value v-node f-node value)))
  (when (= (acknowledge-count f-node) 0) 
    (acknowledge *caller*)))				    
	       
    


#|
(defmethod relax ((f-node function-node))
  (let ((values (apply (relax-func f-node) 
			  (loop for v-node in (value-nodes f-node)
			      collect (read-value  v-node)))))
    (if (loop for v-node in (value-nodes f-node)
	    for new-value in values
	    for old-value in (loop for v-node in (value-nodes f-node)
				 do (lock v-node)
				 collect (read-value  v-node))
	    always (smaller-or-equal new-value old-value))
	(loop for v-node in (value-nodes f-node)
	    for new-value in values	       
	    do (write-value v-node f-node new-value)
	    do (unlock v-node))
      (relax f-node))))
|#



			  
(defpargeneric relax-net :now (net))

(defpargeneric relax-part :future (net))



(defmethod stop-relax ((net relaxation-net))
  (setf (stopped net) t))

(defmethod relax-net ((net relaxation-net))
 (multiple-value-bind (sec min hour) 
     (get-decoded-time) 
   (format t "~a:~a,~a Uhr" hour min sec))
  (mapcar #'touch (loop for part in (partitions net)
		       collect (prog1 (relax-part part) (sleep 0.1)))))

(defmethod relax-part ((part partition))
  (loop for f in (function-nodes part)
      do  (progn (incf (acknowledge-count part))
		 (relax f)))
  (loop until (= (acknowledge-count part) 0)
      do (wait-for (acknowledge))))



(defmethod acknowledge ((part partition))
  (decf (acknowledge-count part)))
  

(let ((spaces ()))
  (defun my-make-instance (class-name &rest initargs)
    (if (endp spaces)
	(progn (setf spaces (spaces *manager*))
	       (apply #'make-instance class-name initargs))
      (prog1 (move (apply #'make-instance class-name initargs) 
		   (first spaces)) 
	(setf spaces (rest spaces))))))
      
     
			   

#|
(defmethod value-ok ((node value-node) value)
  (declare (ignore value))
  t)

(defmethod value-new ((node value-node) value)
  (> (abs (- value (content node))) 0.4))

(defun ari-medium (&rest numbers)
  (let* ((x (length numbers))
	 (y (/ (apply #'+ (mapcar #'touch numbers))
	       x)))
    (make-list x :initial-element y)))

(defun geo-medium (&rest numbers)
  (let* ((x (length numbers))
	 (y (expt (apply #'* (mapcar #'touch numbers))
		  (/ 1 x))))
    (make-list x :initial-element y)))

(defun all-numberp (&rest items)
  (loop for item in items
      always (numberp (touch item))))




(make-net (list (list :relax-func (fdefinition 'geo-medium) :active-pred (fdefinition 'all-numberp))
		(list :relax-func (fdefinition 'ari-medium) :active-pred (fdefinition 'all-numberp)))
	  '((:content 1) (:content 12) (:content 3) (:content 15) (:content 7))
	  '((1 0 0 1 1)
	    (0 1 1 1 0)))

(compile-graph relaxation-net ((function-node wqer) (value-node zook)) ((0 1) (1 0)) ("ki19" "ki20"))


(defnet test 
    ((a :content 1) (b :content 12) (c :content 3) (d :content 15) (e :content 7))
  ((v1 #'geo-medium a d e) (v2 #'ari-medium b c d) (v3 #'ari-medium a b c e)))
|#


(defpargeneric logg :now (obj id &optional comment))

(defpargeneric get-logg :now (obj id))

(defpargeneric display-logg :now (obj))

(defpargeneric init-logg :now (obj &rest ids))

(defvar *logging-on* nil)

(defmethod logg ((obj logg-obj) identifier &optional comment)
  (when *logging-on*
    (if (gethash identifier (logg-table obj))
	(progn (incf (car (gethash identifier (logg-table obj))))
	       (push comment (cdr (gethash identifier (logg-table obj)))))
      (setf (gethash identifier (logg-table obj)) (list 1 comment)))))

(defmethod get-logg ((obj logg-obj) identifier)
  (or (gethash identifier (logg-table obj))
      '(0)))

(defmethod init-logg  ((obj logg-obj) &rest ids)
  (if ids
      (loop for id in ids do (remhash id (logg-table obj)))
    (setf (logg-table obj) (make-hash-table)))
  ())
    

(defmethod display-logg ((obj logg-obj))
  (format t "Logg for ~a ~a:~%" (class-name (class-of obj)) (name obj))
  (maphash #'(lambda (key value)
	       (format t "    logg for ~a is ~a.~%" key value))
	   (logg-table obj)))

(defmethod relax :before ((obj logg-obj))
  (logg obj 'relaxing-start (multiple-value-bind (sec min hour) 
				(get-decoded-time) 
			      (format nil "~a:~a,~a Uhr" hour min sec))))

(defmethod relax :after ((obj logg-obj))
  (logg obj 'relaxing-end (multiple-value-bind (sec min hour) 
				(get-decoded-time) 
			    (format nil "~a:~a,~a Uhr" hour min sec))))

(defmethod acknowledge :before ((obj logg-obj))
  (logg obj 'acknowledge))

(defmethod read-value :before ((obj logg-obj))
  (when (mp:process-lock-locker (v-lock obj))
    (logg obj 'access-when-locked))
  (logg obj 'read (multiple-value-bind (sec min hour) 
		      (get-decoded-time) 
		    (format nil "~a:~a,~a Uhr" hour min sec))))



(defmethod write-value :before ((obj logg-obj) writer value)
  (logg obj 'write (multiple-value-bind (sec min hour) 
		       (get-decoded-time) 
		     (format nil "~a:~a,~a Uhr" hour min sec))))


(defmethod receive-call :before (met c (obj logg-obj) args)
  ())
;;;(let ((*current-actor* obj)) (unless (member met '(get-logg display-logg)))
;;;				 (logg obj (cons 'remote-access met))
     
(defmethod semi-border-node ((node function-node))
  (and (not (border-node node))
       (loop for task in (connected-tasks node)
	   thereis (border-node task))))

(defmethod semi-border-node ((node value-node)) ())

(defmethod border-node ((node net-node))
  (not (loop for task in (connected-tasks node)
	   always (eq (location node)
		      (location task)))))
		    
(defmethod border-node ((obj t)) t)
(defmethod name ((obj t)) "no name")

(defmethod name ((obj net-node)) (node-name obj))


		      




