;;;; -*- Mode: LISP; Syntax: Common-lisp;  Package: nc; Base: 10 -*-
;;;
;;; TRACE CHANNELS
;;;
;;; fb 1.3.94, 2.3.94
;;;
;;; TODO: (Unklarheiten sind (falls klar, dass unklar) mit ### markiert)

(in-package :nc)

#|

  Trace-Channels
  -------------------
  Zweck:  TCs geben Entwicklern von einzelnen Modulen oder Programmteilen 
	  die Moeglichkeit, Debugging-Information zur Entwicklungszeit
	  kontrolliert auszugeben.
          
  Hintergrund:
          TCs ersetzen die bewaehrte (aber konzeptionell unstrukturierte) 
          Ausgabe von Debugging-Information via (format ...)  auf 
          *standard-output* oder speziellere Streams. 
          Hierbei ist insbesondere nachteilig, dass diese Ausgaben
          nicht einfach zur Laufzeit ein- oder ausgeschaltet werden koennen,
          und dass bei Ausgaben von mehreren Modulen auf einen Stream nicht
          ohne weiteres dynamisch vom Benutzer selektiert werden kann, welche Module
          etwas ausgeben sollen oder nicht.

  Beispiel fuer Verwendung:

          (defparameter *my-trace-channel* (make-trace-channel :foobarbaz))
          ;;...
          (defun foo (bar)
	    (format-trace *my-trace-channel* "Now entering Foo, is this ok?...")
	    ;;...
	    (when (trace-channel-enabled-p *my-trace-channel*)
	      (let ((baz (do-some-heavy-computation-just-for-tracing bar)))
	         (format-trace *my-trace-channel* 
	                       "Value of argument Bar is ~S, its cost is ~S" bar baz)))
            ;;...
	    )

	  
  Benutzersicht:
          TCs werden von der Benutzeroberflaeche je nach ihrer Beschaffenheit
          (graphisch, terminal-basiert, etc.) unterschiedlich realisiert.
          
          In einer graphischen Benutzeroberflaeche gibt es mindestens ein
          Fenster, in dem vom Benutzer auswaehlbare TCs angezeigt werden.

          In einer "rohen" Terminaloberflaeche erscheinen die TC Ausgaben
          im laufenden Ausgabetext.
                    
  Externes Protokoll (fuer Client-Module):

    Definitionen:

	MAKE-TRACE-CHANNEL 	                 	[FUNKTION]
          name &key (enabled-p t)

	FORMAT-TRACE					[GENERISCHE FUNKTION]
          trace-channel format-string format-args

	TRACE-CHANNEL-ENABLED-P				[GENERISCHE FUNKTION]
	  trace-channel
          
  Spezialisierungsprotokoll:

     Definitionen:

        STANDARD-TRACE-MANAGER				[KLASSE]

	MAKE-TRACE-MANAGER-FROM-OLD 			[GENERISCHE FUNKTION]
	   old-tm new-manager-class-name
	   
	INSTALL-STANDARD-TRACE-MANAGER                  [FUNKTION]   
	   
	UI-INSTALL-TRACE-MANAGER			[GENERISCHE FUNKTION]
           user-interface

	UI-TRACE-MANAGER-CLASS-NAME 			[GENERISCHE FUNKTION]	
	  user-interface

	INSTALL-NEW-TRACE-MANAGER			[GENERISCHE FUNKTION]
	  old-tm new-tm

	TRACE-MANAGER-FORMAT				[GENERISCHE FUNKTION]
	  tm tc format-string format-args
	
	TRACE-MANAGER-ADD-CHANNEL			[GENERISCHE FUNKTION]
	  trace-manager channel

	MAP-OVER-ALL-CHANNELS				[GENERISCHE FUNKTION]
	 function trace-manager

	GET-ALL-TRACE-MANAGER-CHANNELS			[GENERISCHE FUNKTION]
	  trace-manager

	FIND-TRACE-CHANNEL				[GENERISCHE FUNKTION]
	  trace-manager channel-name
	  
  |#


(defvar *trace-manager* nil 
  "Der momentan aktive Trace Manager. 
   Zugriff ueber CURRENT-TRACE-MANAGER und SET-CURRENT-TRACE-MANAGER")

(defun current-trace-manager ()  
  "Der momentan aktive Trace Manager."
  *trace-manager*)

(defun set-current-trace-manager (tm) 
  "Installieren eines neuen aktiven Trace Managers."
  (check-type tm standard-trace-manager)
  (setf *trace-manager* tm))

(defun set-current-trace-manager-stream (stream) 
  "Setzen des Ausgabestreams fuer einen trace manager"
  (setf (stream *trace-manager*) stream))

(defclass standard-trace-channel ()
  ((name 
    :type symbol
    :reader trace-channel-name
    :initform nil
    :initarg :name
    :documentation "Name des TC zur Identifikation durch den Benutzer.")
   (stream
    :accessor stream
    :initform nil
    :initarg :stream
    :documentation 
    "Stream auf den der Channel schreibt. Wenn nil, dann wird
     der Stream des trace managers verwendet")
   (enabled-p
    :type (member t nil)
    :accessor trace-channel-enabled-p
    :initform t
    :initarg :enabled-p
    :documentation 
      "True, wenn der Trace Channel aktiviert ist, d.h.
       Ausgaben auch angezeigt werden.
       Der Client kann mit TRACE-CHANNEL-ENABLED-P erfragen, ob ein
       TC ueberhaupt gerade angezeigt wird. Auf diese Weise kann er
       eventuell allein fuer die Ausgabe erforderliche Berechnungen
       einsparen. Der Status eines TCs (aktiviert oder nicht) wird
       vom Trace Channel Manager bestimmt.")
   ;; messages with equal or higher priority signal an error
   (break-priority :accessor break-priority
		   :initarg :break-priority)
   ;; messages with priority equal or higher are traceged
   (trace-priority :accessor trace-priority
		   :initarg :trace-priority)))
   

(defclass standard-trace-manager ()
  ((channels
    :type hash-table
    :accessor trace-manager-channels
    :initform (make-hash-table :test #'eq)
    :initarg :trace-channels
    :documentation "Hash-Table, der Namen auf Trace-Channels abbildet.")
   (stream
    :accessor stream
    :initarg :stream
    :initform *terminal-io*)
   (new-initargs
    :type list
    :accessor trace-manager-new-initargs
    :initform nil
    :documentation "Dieser Slot wird gebraucht, um zu umgehen, dass 
     CHANGE-CLASS keine
     Initargs akzeptiert. Siehe MAKE-TRACE-MANAGER-FROM-OLD."))
  (:documentation
   " Ein Trace Manager verwaltet alle Trace Channels, die erzeugt werden. 
     Ausgaben ueber FORMAT-TRACE auf TCs werden an diesen deligiert.
     Genau ein Exemplar dieser Klasse wird bei der allerersten Instantiierung
     irgendeines TCs durch MAKE-TRACE-CHANNEL erzeugt und mittels 
     SET-CURRENT-TRACE-MANAGER abgelegt. Es ist dann mit CURRENT-TRACE-MANAGER
     zugreifbar. 
     Diese Klasse soll von speziellen Oberflaechen spezialisiert werden,
     um die Praesentation der Trace-Information der Oberflaeche anzupassen."))


(defun make-trace-channel (name &key (enabled-p t)
				     (intern-keyword-p t)
				     (redifinition-warning-p nil)
				     (class 'standard-trace-channel)
				     (trace-priority (priority :note))
				     (break-priority (priority :err)))
  "Benoetigt ein bestimmtes Modul einen oder mehrere TCs, so erzeugt 
   es diese zunaechst mittels MAKE-TRACE-CHANNEL. 
   Dieses kann schon zur Ladezeit des Moduls erfolgen. (Natuerlich nur,
   wenn bereits dieses File geladen wurde.)
   Ist noch keine Benutzeroberflaeche geladen oder gestartet,
   so erfolgen Ausgaben auf den TC auf *terminal-io*.
   Ist schon eine spezielle Benutzeroberlaeche geladen und
   gestartet worden, so erscheinen die Ausgaben in den TC in benutzeroberflaechen-
   spezifischer Weise. (Siehe Spezialisierungsprotokoll fuer Trace Manager)
   Resultat ist ein neuer Trace Channel.
   <name> ist ein Symbol. Stammt dieses nicht aus der Keyword-Package, so
   wird ein gleichnamiges Symbol in der Keyword-Package als Name benutzt, falls
   <intern-keyword-p> true ist.
   <enabled-p> setzt den Status des Trace-Channels. Deafault ist true, also
   'angeschaltet'.
   Existiert bereits ein Trace Channel mit dem gleichen Namen, so wird
   dieser ueberschrieben. Falls <redifinition-warning-p> true ist, wird dann
   ein Warning ausgegeben."

  
  (unless (priority-p trace-priority)
    (setq trace-priority (priority trace-priority)))
  
  (unless (priority-p break-priority)
    (setq break-priority (priority break-priority)))
  
  (when intern-keyword-p
    (setq name (intern 
                (when (symbolp name) (symbol-name name))
                (find-package :keyword))))
    
  (unless (current-trace-manager)
    (set-current-trace-manager 
     (make-trace-manager-from-old nil 'standard-trace-manager)))
  
  (let ((tm (current-trace-manager)))

    (when (and redifinition-warning-p (find-trace-channel tm name))
      (warn "Trace-Channel ~S wird ueberdefiniert" name)) ;### auf TC "intern" ausgeben

    (trace-manager-add-channel tm 
			       (make-instance class
				 :name name
				 :enabled-p enabled-p
				 :trace-priority trace-priority
				 :break-priority break-priority))))

(defgeneric format-trace (trace-channel format-string &rest format-args)
  (:documentation 
   "Ausgabe wie mit format auf trace-channel.
    Eine Ausgabe erfolgt nur, wenn der Trace Channel aktiviert ist, d.h.
    (trace-channel-enabled-p trace-channel) ist true.
    Der Wert dieser Funktion ist nicht spezifiziert."))


;;; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
;;; Ab hier Generische Funktionen des Spezialisierungsprotokolls
;;; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

(defgeneric make-trace-manager-from-old (old-tm new-manager-class-name &rest initargs)
  (:documentation
   "Erzeugen eines neuen Trace Channel Managers, wobei die Slots
    des alten TM weitestgehend uebernommen werden.
    Dieses ist wichtig, wenn schon beim Laden mit MAKE-TRACE-CHANNEL
    TCs und damit ein TM erzeugt werden, der aber noch nicht speziell genug
    fuer die spaeter zu ladende Oberflaeche sein kann."))

(defgeneric ui-install-trace-manager (ui)
  (:documentation 
   "Muss von jeder Oberflaeche beim Start aufgerufen werden.
    Kann fuer spezielle Oberflaechen spezialisiert werden, um den jeweiligen TM 
    dieser Oberflaeche zu installieren. (Doch siehe UI-TRACE-MANAGER-CLASS-NAME.)"))

(defgeneric ui-trace-manager-class-name (ui)
  (:documentation 
   "Name (symbol) der Klasse, die als Trace-Manager von UI-INSTALL-TRACE-MANAGER
    instantiiert und installiert werden soll. Muss eine Unterklasse von 
    STANDARD-TRACE-MANAGER benennen. Speziellere Oberflaechen sollen diese
    Funktion spezialisieren, um ihren eigenen Trace Manager einzusetzen.
    Wird diese Funktion spezialisiert, so muss nicht UI-INSTALL-TRACE-MANAGER
    spezialisiert werden."))

(defgeneric install-new-trace-manager (old-tm new-tm)
  (:documentation 
   "In Abhaengigkeit vom vorher installierten Trace-Manager <old-tm> und dem neu
    geforderten <new-tm> werden eventuell Aufraeum- und Umstelltaetigkeiten durchgefuehrt,
    und als Seiteneffekt mittels SET-CURRENT-TRACE-MANAGER der <new-tm> installiert."))

(defgeneric trace-manager-format (tm tc prio format-string format-args)
  (:documentation 
   "An diese Funktion werden Ausgaben auf Trace Channels mittels FORMAT-LOG deligiert."))

(defgeneric trace-manager-add-channel (trace-manager channel)
  (:documentation 
   "Fuegt einen Trace-Channel zu den von Trace Manager verwalteten Channels
    hinzu. Existiert bereits ein TC mit gleichem (eql) Namen, so wird dieser ersetzt."))
   
(defgeneric map-over-all-channels (function trace-manager)
  (:documentation 
   "Wendet <function> (zwei Argumente, Name und Trace-Channel) auf alle 
    (Name Trace-Channel)-Paare an."))

(defgeneric get-all-trace-manager-channels (trace-manager)
  (:documentation "Ausgabe aller vom Trace Manager verwalteten Trace Channels als Liste."))

(defgeneric find-trace-channel (trace-manager channel)
  (:documentation 
   "Gibt den Trace-Channel mit Namen (Symbol) <channel> zurueck, wenn er von <trace-manager>
    verwaltet wird. Ist <channel> ein Trace-Channel, so wird nur <channel> zurueckgegeben,
    wenn er auch von <trace-manager> unter seinem Namen verwaltet wird.
    Falls nicht gefunden Rueckgabewert nil."))
    

;;; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
;;; IMPLEMENTATION
;;; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 



(defmethod initialize-instance :after ((tm standard-trace-manager) &key &allow-other-keys)
  "Macht nichts ausser &key &allow-other-keys.")


(defun install-standard-trace-manager ()
  (install-new-trace-manager 
   (current-trace-manager) 
   (make-trace-manager-from-old (current-trace-manager) 
				'standard-trace-manager)))





;;; MAKE-TRACE-MANAGER-FROM-OLD

(defmethod make-trace-manager-from-old ((old-tm null) new-manager-class-name &rest initargs)
  "Ohne Vorgaenger old-tm ein normaler Konstruktor fuer TMs."
    (apply #'make-instance new-manager-class-name initargs))

(defmethod make-trace-manager-from-old ((old-tm standard-trace-manager) 
					new-manager-class-name &rest initargs)
  "Neuen Trace Manager erzeugen.
   Wenn vorher schon ein standard-trace-manager installiert war, seine Trace-
   Channels uebergeben."
  ;; schoen waer's doch keine initargs in der Def. von change-class vorhanden...
  ;; das geht also nicht -> (apply #'change-class old-tm new-manager-class-name initargs)
  ;; ...
  ;; Daher speichern wird die initargs erstmal und "umrunden"
  ;; UPDATE-INSTANCE-FOR-DIFFERENT-CLASS (siehe unten)
  ;;
  ;; Problem:   install-new-trace-manager bekommt dann nicht mehr den alten tm zum dispatchen!
  ;;(setf (trace-manager-new-initargs old-tm) initargs)
  ;;(change-class old-tm new-manager-class-name) ;ruft UPDATE-INSTANCE-FOR-DIFFERENT-CLASS auf
  
  (apply #'make-instance 
	 new-manager-class-name
	 :trace-channels (trace-manager-channels old-tm) ;Hash-table wird weitergegeben
	 initargs)
  )

#|
(defmethod update-instance-for-different-class :around ((old standard-trace-manager) 
							(new standard-trace-manager) &rest initargs)
  "Ruft (call-next-method) mit den in einem Slot von <old> durch MAKE-TRACE-MANAGER-FROM-OLD 
   gespeicherten initargs auf. Auf diese Weise wird die traurige Tatsache umgangen, dass in 
   CLOS CHANGE-CLASS keine initargs annimmt, um sie zu UPDATE-INSTANCE-FOR-DIFFERENT-CLASS durchzureichen.
   (siehe auch MOP pp. 32)."
  (declare (ignore initargs))
  (apply #'call-next-method old new (trace-manager-new-initargs old)))
|#						

;;; INSTALL-NEW-TRACE-MANAGER

(defmethod install-new-trace-manager ((old-tm null) (new-tm standard-trace-manager))
  (set-current-trace-manager new-tm))

(defmethod install-new-trace-manager ((old-tm standard-trace-manager)
				      (new-tm standard-trace-manager))
  (set-current-trace-manager new-tm))



;;; FORMAT-TRACE (Ausgabe)

;; format-trace wirkt ueberhaupt nur, wenn TRACE-CHANNEL-ENABLED-P
;; Dieses Verhalten kann schon hier mittels :around Methode ausgedrueckt werden.

(defmethod format-log :around ((tc standard-trace-channel)
			       (prio number)
			       format-string &rest format-args)
  (declare (ignore format-string format-args))
  (when (and (trace-channel-enabled-p tc)
	     (>= prio (trace-priority tc)))
    (call-next-method))
  (when (>= prio (break-priority tc))
    (break "Trace-Channel-induced break"))
  t)

(defmethod format-log (tc (prio-name symbol) format-string &rest  format-args)
  (apply #'format-log tc (priority prio-name) format-string format-args))
    

(defmethod format-trace ((tc standard-trace-channel) format-string &rest format-args)
  "format-trace is a shortcut for loggin' a message with priority note"
  (assert (not (null (current-trace-manager))))
  (apply #'format-log 
   ;; (current-trace-manager)
	      tc
	      (priority :note)
	      format-string 
	      format-args)
  (values))

(defmethod format-trace ((tc symbol) format-string &rest format-args)
  "format-trace is a shortcut for loggin' a message with priority note"
  (assert (not (null (current-trace-manager))))
  (format-log 
   ;; (current-trace-manager)
	      tc
	      (priority :note)
	      format-string 
	      format-args)
  (values))

 
(defmethod format-log ((tc standard-trace-channel) (prio number) format-string &rest format-args)
  (assert (not (null (current-trace-manager))))
  (trace-manager-format
   (current-trace-manager)
   tc
   prio 
   format-string 
   format-args)
  (values))




(defmethod format-log ((tc-name symbol) prio format-string &rest format-args) 
  "Wenn als Trace Channel ein Symbol angegeben wird, wird dieses
   als Name eines Trace-Channels aufgefasst und ganz normal verfahren."
  (if (null (current-trace-manager))
      ;; ohne Trace-Manager kann das auch kein gueltiger Trace-Channel sein
      ;; (weil nicht sein kann...)
      (warn-no-tc tc-name format-string format-args)
    (let ((tc (find-trace-channel (current-trace-manager) tc-name)))
      (if (null tc)
	  (warn-no-tc tc-name format-string format-args)
	(apply #'format-log tc prio format-string format-args)))))


;; Manchmal vergisst man das make-trace-channel...

(defmethod format-log ((tc t) prio format-string &rest format-args) 
  (declare (ignore prio))
  (warn-no-tc tc format-string format-args))

(defmethod format-log ((tc null) prio format-string &rest format-args) 
  (declare (ignore prio))
  (warn-no-tc tc format-string format-args))


;;; ### Dieses als generische Funktion des Trace Managers? 
;;; (Beim CLIM-Tracer koennten dann die Warnings trotzdem im CLIM-Frame erscheinen...)

(defun  warn-no-tc (tc format-string format-args)
  ;; NICHT konwerk-warn, denn falls konwerk-warn auf einen nicht vorhandenen 
  ;; trace-channel geleitet wird -> never ending story
  (warn "Ausgabe auf nicht vorhandenen Trace Channel ~S ~A" 
	tc (apply #'format nil format-string format-args))
  (values))



;;; TRACE-MANAGER-FORMAT


#+:ALLEGRO(defvar *trace-stream* excl:*initial-terminal-io*)
#-:ALLEGRO(defvar *trace-stream* *terminal-io*)


(defmethod trace-manager-format ((tm standard-trace-manager) tc prio format-string format-args)
  "Standard-Trace-Manager leitet alle Ausgaben auf *terminal-io* um."
  (apply #'format 
	 *trace-stream*
	  (concatenate 'string "~&~A ~A: ~@<" format-string "~:>")
	  (symbol-name (trace-channel-name tc))
	  (symbol-name (priority-name prio))
	  format-args)
  (values))



;;; Zugriff auf die Channels eines TM


(defmethod set-channel-active ((tc-name symbol))
  (let ((tc (find-trace-channel (current-trace-manager) tc-name)))
    (assert tc)
    (setf (trace-channel-enabled-p tc) t)))

(defmethod set-channel-trace-priority ((tc-name symbol) prio)
  (let ((tc (find-trace-channel (current-trace-manager) tc-name)))
    (assert tc)
    (setf (trace-priority tc) (priority prio))))

(defmethod set-channel-inactive ((tc-name symbol))
  (let ((tc (find-trace-channel (current-trace-manager) tc-name)))
    (assert tc)
    (setf (trace-channel-enabled-p tc) nil)))


(defmethod set-channel-stream ((tc-name symbol) stream)
  (let* ((tc (find-trace-channel (current-trace-manager) tc-name))
	 (old-stream (stream tc)))
    (assert tc)
    (setf (stream tc) stream)
    old-stream))

(defmethod trace-manager-add-channel ((tm standard-trace-manager) channel)
  (setf (gethash (trace-channel-name channel) 
		 (trace-manager-channels tm))
    channel))

(defmethod map-over-all-channels (function (tm standard-trace-manager))
  (maphash function (trace-manager-channels tm)))

(defmethod get-all-trace-manager-channels ((tm standard-trace-manager))
  (loop for channel being each hash-value in (trace-manager-channels tm)
	collect channel))

(defmethod find-trace-channel ((tm standard-trace-manager) (channel-name symbol))
  (gethash channel-name (trace-manager-channels tm)))

(defmethod find-trace-channel ((tm standard-trace-manager) (channel standard-trace-channel))
  (and (eq (find-trace-channel tm (trace-channel-name channel))
	   channel)
       channel))

;;; TC PRAEDIKAT

(defun trace-channel-p (obj)
  "True g.d.w <obj> ein STANDARD-TRACE-CHANNEL ist, sonst NIL."
  (typep obj 'standard-trace-channel))


(defmethod print-object ((object standard-trace-channel) stream)
  (print-unreadable-object (object stream :type t :identity t)
    (format stream "~S " (trace-channel-name object)))
  object)





;;;; erweiterungen 

(defvar *priority-names* '(:debug :info :note :warn :err :crit :emerg))
(defvar *max-prio* (length *priority-names*))

(defun priority (priority-name &optional (errorp t))
  (let ((pos (position priority-name *priority-names*)))
    (if pos pos
      (when errorp (error "Unknown priority-name ~S" priority-name)))))

(defun priority-name (priority &optional (errorp t))
  (let ((name (nth priority *priority-names*)))
    (if name name
      (when errorp (error "Priority ~S does not exist" priority)))))

(defun priority-p (prio)
  (and (integerp prio)
       (<= 0 prio *max-prio*)))
